(function() {
  var bigint, types, util;

  bigint = require("bigint");

  util = require("util");

  types = require("../types");

  exports.serialize = function(record, buffer, offset) {
    var advance, append, ensure, fields, grow;
    if (buffer == null) buffer = new Buffer(1);
    if (offset == null) offset = 0;
    fields = record.constructor.fields;
    if (!(fields != null)) {
      throw new TypeError("" + (util.inspect(record)) + "# doesn't look like a Record");
    }
    grow = function() {
      var resized;
      resized = new Buffer(buffer.length * 2);
      buffer.copy(resized);
      return buffer = resized;
    };
    ensure = function(bytes) {
      var _results;
      _results = [];
      while (offset + bytes >= buffer.length) {
        _results.push(grow());
      }
      return _results;
    };
    advance = function(bytes) {
      var currentOffset, _ref;
      _ref = [offset, offset + bytes], currentOffset = _ref[0], offset = _ref[1];
      return currentOffset;
    };
    append = function(type, value) {
      var bodyOffset, bytesWritten, charsWritten, element, elementType, name, _i, _len, _ref, _results, _results2;
      if (type.fields) {
        _ref = type.fields;
        _results = [];
        for (name in _ref) {
          elementType = _ref[name];
          _results.push(append(elementType, elementType.normalize(value[name])));
        }
        return _results;
      } else if (type === types.byte) {
        ensure(1);
        return buffer.writeUInt8(value, advance(1));
      } else if (type === types.boolean) {
        ensure(1);
        return buffer.writeUInt8((value ? 1 : 0), advance(1));
      } else if (type === types.int) {
        ensure(4);
        return buffer.writeInt32BE(value, advance(4));
      } else if (type === types.long) {
        ensure(8);
        buffer.fill(0, offset, offset + 8);
        return value.toBuffer({
          size: 8
        }).copy(buffer, advance(8));
      } else if (type === types.float) {
        ensure(4);
        return buffer.writeFloatBE(value, advance(4));
      } else if (type === types.double) {
        ensure(8);
        return buffer.writeDoubleBE(value, advance(8));
      } else if (type === types.ustring) {
        if (value != null) {
          ensure(4);
          bodyOffset = offset + 4;
          bytesWritten = buffer.write(value, bodyOffset);
          charsWritten = Buffer._charsWritten;
          if (Buffer._charsWritten < value.length) {
            ensure(Buffer.byteLength(value));
            bytesWritten = buffer.write(value, bodyOffset);
          }
        } else {
          bytesWritten = -1;
        }
        ensure(4);
        buffer.writeInt32BE(bytesWritten, advance(4));
        if (bytesWritten > 0) return advance(bytesWritten);
      } else if (type === types.buffer) {
        if (value != null) {
          ensure(value.length + 4);
          buffer.writeInt32BE(value.length, advance(4));
          return value.copy(buffer, advance(value.length));
        } else {
          ensure(4);
          return buffer.writeInt32BE(-1, advance(4));
        }
      } else if (type instanceof types.VectorType) {
        if (value != null) {
          ensure(4);
          buffer.writeInt32BE(value.length, advance(4));
          _results2 = [];
          for (_i = 0, _len = value.length; _i < _len; _i++) {
            element = value[_i];
            _results2.push(append(type.elementType, type.elementType.normalize(element)));
          }
          return _results2;
        } else {
          ensure(4);
          return buffer.writeInt32BE(-1, advance(4));
        }
      }
    };
    append(record.constructor, record);
    return buffer.slice(0, offset);
  };

  exports.deserialize = function(type, buffer, offset) {
    var advance, fields, longBuffer, unpack;
    if (offset == null) offset = 0;
    fields = type.fields;
    longBuffer = null;
    advance = function(bytes) {
      var currentOffset, _ref;
      _ref = [offset, offset + bytes], currentOffset = _ref[0], offset = _ref[1];
      return currentOffset;
    };
    unpack = function(type) {
      var elementType, name, readWithLength, record, start, _ref;
      readWithLength = function(reader) {
        var length;
        length = buffer.readInt32BE(advance(4));
        if (length < 0) {
          return null;
        } else {
          return reader(length);
        }
      };
      if (type.fields) {
        record = new type;
        _ref = type.fields;
        for (name in _ref) {
          elementType = _ref[name];
          record[name] = unpack(elementType);
        }
        return record;
      } else if (type === types.byte) {
        return buffer.readUInt8(advance(1));
      } else if (type === types.boolean) {
        return !!buffer.readUInt8(advance(1));
      } else if (type === types.int) {
        return buffer.readInt32BE(advance(4));
      } else if (type === types.long) {
        if (longBuffer == null) longBuffer = new Buffer(8);
        start = advance(8);
        buffer.copy(longBuffer, 0, start, offset);
        return bigint.fromBuffer(longBuffer, {
          size: 8
        });
      } else if (type === types.float) {
        return buffer.readFloatBE(advance(4));
      } else if (type === types.double) {
        return buffer.readDoubleBE(advance(8));
      } else if (type === types.ustring) {
        return readWithLength(function(length) {
          start = advance(length);
          return buffer.toString("utf8", start, offset);
        });
      } else if (type === types.buffer) {
        return readWithLength(function(length) {
          var valueBuffer;
          valueBuffer = new Buffer(length);
          start = advance(length);
          buffer.copy(valueBuffer, 0, start, offset);
          return valueBuffer;
        });
      } else if (type instanceof types.VectorType) {
        return readWithLength(function(length) {
          var i, _results;
          _results = [];
          for (i = 0; 0 <= length ? i < length : i > length; 0 <= length ? i++ : i--) {
            _results.push(unpack(type.elementType));
          }
          return _results;
        });
      } else {
        throw new Error("Unknown type: " + (util.inspect(type)));
      }
    };
    return unpack(type);
  };

}).call(this);
